{
*****************************************************************************
*  Name    : USART4 Library                                                 *
*  Author  : Jerry Messina (based on module by David John Barker            *
*  Notice  : Copyright (c) 2003 Mecanique                                   *
*          : All Rights Reserved                                            *
*  Date    : 06 MAY 2023                                                    *
*  Version : 2.3 JM  06 MAY 2023                                            *
*          : add support for Q83/Q84 (USART_XV18CORE)                       *
*  Version : 2.2 JM  08 NOV 2022                                            *
*          : add '#option USART4_TXEN' to skip writebyte if TX is disabled  *
*          : add sub TXEnable() to enable/disable TX at runtime             *
*          : add br250000 for clock >= 16                                   *
*  Version : 1.9 JM                                                         *
*          : correct register map for U4 registers (_xv18)                  *
*          : add support for Q43 family                                     *
*          : change REPEAT_UNTIL_BIT() macro to use device file '_access_x' *
*          : change ReadByte() to use REPEAT_UNTIL_BIT()                    *
*  Version : 1.8 JM                                                         *
*          : based on USART2 v1.8  10/16/2019                               *
*          : USART4 supported by J94/J99, 6xK40, and Q43 (future)           *
*****************************************************************************
}
module USART4

include "system.bas"

// check if device supports USART4...
#if (_usart < 4)
  #error _device + " does not support USART4"
#endif

// calculate clock...
const OSC = _clock
const FOSC = OSC * 1000000

// xv18 core devices added v1.8
// this option is set automatically based on the device file '_xv18' setting
#if IsOption(USART_XV18CORE) and not (USART_XV18CORE in (true, false))
  #error USART_XV18CORE, "Invalid option. USART_XV18CORE must be TRUE or FALSE"
#endif
#if (_xv18 = $01)
  #option USART_XV18CORE = true
#else
  #option USART_XV18CORE = false
#endif

// low speed or high speed (BRGH)...
#if IsOption(USART4_BRGH) and not (USART4_BRGH in (true, false))
  #error USART4_BRGH, "Invalid option. BRGH must be TRUE or FALSE"
#endif
#option USART4_BRGH = true

// 8 bit or 16 bit BRG (BRG16 only supported with EUSART modules)...
#if IsOption(USART4_BRG16) and not (USART4_BRG16 in (true, false))
  #error USART4_BRG16, "Invalid option. BRG16 must be TRUE or FALSE"
#endif
#if (USART_XV18CORE)                        // xv18 devices require BRG16
  #option USART4_BRG16 = true
  #if (USART4_BRG16 = false)
    #error USART4_BRG16, "Invalid option. xv18 requires BRG16 = TRUE"
  #endif
#endif
#option USART4_BRG16 = false
#if not USART4_BRG16 and not USART4_BRGH    // BRG16 = 0, BRGH = 0
  const FMULT = 64
#elseif USART4_BRG16 and not USART4_BRGH    // BRG16 = 1, BRGH = 0
  const FMULT = 16
#elseif not USART4_BRG16 and USART4_BRGH    // BRG16 = 0, BRGH = 1
  const FMULT = 16
#else                                       // BRG16 = 1, BRGH = 1
  const FMULT = 4
#endif

// OERR handling (added v1.7)
// the USART has a 2-byte input buffer and can easily overflow if bytes are not read 
// often enough. if this occurs OERR will be set and the USART will stop accepting any 
// new characters until the error is cleared.
// setting this option true will automatically clear the error before every read operation, 
// however you will not know that an error has occurred and characters may have been lost
#option USART4_CLEAR_OERR = false

// check TXEN in writebyte, skip if tx disabled (added v2.2)
#option USART4_TXEN = false

//--------------------------------------------------------------------------
// USART4 TX/RX port pins (DEVICE DEPENDANT)
//--------------------------------------------------------------------------
// there aren't really any standard PORT pins for UART4 
// if you are using PPS mapped pins then set these #options to match
// the port pins you're using... you'll still have to setup the PPS yourself
// NOTE: if the pins have analog capability it is up to the user to set 
// the respective pins to digital IO mode (setalldigital.bas)
#if _device in (18F65K40, 18F66K40, 18F67K40) or _
    _device in (18LF65K40, 18LF66K40, 18LF67K40)
  #option USART4_TX = PORTC.0         // USART4 TX output (must be mapped via PPS)
  #option USART4_RX = PORTC.1         // USART4 RX input
#endif  // _device
#if (not isoption(USART4_TX))
  #warning "**VERIFY USART4_TX PIN - assuming PORTD.1**"
  #option USART4_TX = PORTD.1         // USART4 TX output
#endif
#if (not isoption(USART4_RX))
  #warning "**VERIFY USART4_RX PIN - assuming PORTD.0**"
  #option USART4_RX = PORTD.0         // USART4 RX input
#endif
dim USART_TX_PIO as USART4_TX.USART4_TX@,
    USART_RX_PIO as USART4_RX.USART4_RX@

// option to setup usart IO pins at startup
// if true this option adds InitUSARTIO() to the module initialization
// so that the IO pins are set at startup. USART4_TX is normally an idle high
// output but since the pic pin reset state is an input (ie floating), it 
// can be seen as a low by the receiver, generating a serial BREAK, a framing 
// error, or garbage character(s) until it's setup
// this can help, but the real fix is to have a resistor pullup on the TXD pin
#option USART4_STARTUP_INITIO = false

//--------------------------------------------------------------------------
// map registers to USARTx (DEVICE DEPENDANT)
//--------------------------------------------------------------------------
// usart registers for xv18 core devices with 4 usarts
#if (USART_XV18CORE)
public dim
    UxCON0      as U4CON0,
    UxCON1      as U4CON1,
    UxCON2      as U4CON2,
    UxERRIR     as U4ERRIR,
    UxERRIE     as U4ERRIE,
    UxUIR       as U4UIR,
    UxFIFO      as U4FIFO,
    UxBRGL      as U4BRGL,
    UxBRGH      as U4BRGH,
    UxRXB       as U4RXB,
    UxTXB       as U4TXB

public dim
    RCRegister  as UxRXB,
    TXRegister  as UxTXB
    
// alias commonly used public bitnames...
public dim
    BRGS        as UxCON0.7,    // Baud rate Generator Speed Select
    TXEN        as UxCON0.5,    // Transmit Enable Control
    RXEN        as UxCON0.4,    // Receive Enable Control

    SPEN        as UxCON1.7,    // Serial Port Enable (ON bit)
    
    TXMTIF      as UxERRIR.7,   // Transmit Shift Register Empty Interrupt Flag
    FERIF       as UxERRIR.3,   // Framing Error Interrupt Flag
    RXFOIF      as UxERRIR.1    // Receive FIFO Overflow Interrupt Flag

public dim
    BRGH        as BRGS,
    TRMT        as TXMTIF,
    FERR        as FERIF,
    OERR        as RXFOIF
    
// public boolean status flags...
public dim
    Overrun     as UxERRIR.booleans(1),    // RXFOIF, OERR
    FrameError  as UxERRIR.booleans(3)     // FERIF, FERR
    
#else   // standard core device
// usart registers for standard core devices with 4 usarts...

public dim
    SPBRG_Reg   as SPBRG4,
  #if USART4_BRG16
    SPBRGH_Reg  as SPBRGH4,     // added to support 16-bit BRG
    BRG16       as BAUDCON4.bits(3),
  #endif
    RCRegister  as RCREG4,
    TXRegister  as TXREG4,
    RCStatus    as RCSTA4,
    TXStatus    as TXSTA4

// alias public bitnames to TXSTA(x)...
public dim
    CSRC        as TXStatus.7,
    TX9         as TXStatus.6,
    TXEN        as TXStatus.5,
    SYNC        as TXStatus.4,
    BRGH        as TXStatus.2,
    TRMT        as TXStatus.1,
    TX9D        as TXStatus.0

// alias public bitnames to RCSTA(x)...
public dim
    SPEN        as RCStatus.7,
    RX9         as RCStatus.6,
    SREN        as RCStatus.5,
    CREN        as RCStatus.4,
    ADDEN       as RCStatus.3,
    FERR        as RCStatus.2,
    OERR        as RCStatus.1,
    RX9D        as RCStatus.0

// public boolean status flags...
public dim
    ContinousReceive as RCStatus.booleans(4),   // CREN
    Overrun          as RCStatus.booleans(1),   // OERR
    FrameError       as RCStatus.booleans(2)    // FERR
#endif  // USART_XV18CORE

//--------------------------------------------------------------------------
// UART4 interrupt registers and interrupt bits (DEVICE DEPENDANT)
//--------------------------------------------------------------------------
#if (USART_XV18CORE)
// UART4 interrupt registers
dim
    UART_PIR    as PIR12,
    UART_PIE    as PIE12,
    UART_IPR    as IPR12

// UART4 interrupt bit numbers
const
    UART_RCIF = 0,
    UART_TXIF = 1,
    UART_RCIE = UART_RCIF,
    UART_TXIE = UART_TXIF,
    UART_RCIP = UART_RCIF,
    UART_TXIP = UART_TXIF

// J94/J99 devices
#elseif _device in (18F65J94, 18F85J94, 18F95J94, 18F66J94, 18F86J94, 18F96J94, 18F67J94, 18F87J94, 18F97J94) or _
        _device in (18F66J99, 18F86J99, 18F96J99)
// UART4 interrupt registers
dim
    UART_PIR    as PIR6,
    UART_PIE    as PIE6,
    UART_IPR    as IPR6

// UART4 interrupt bit numbers
const
    UART_RCIF = 7,
    UART_TXIF = 6,
    UART_RCIE = UART_RCIF,
    UART_TXIE = UART_TXIF,
    UART_RCIP = UART_RCIF,
    UART_TXIP = UART_TXIF

// 6xK40 devices
#elseif _device in (18F65K40, 18F66K40, 18F67K40) or _
        _device in (18LF65K40, 18LF66K40, 18LF67K40)
// UART4 interrupt registers
dim
    UART_PIR    as PIR4,
    UART_PIE    as PIE4,
    UART_IPR    as IPR4

// UART4 interrupt bit numbers
const
    UART_RCIF = 3,
    UART_TXIF = 2,
    UART_RCIE = UART_RCIF,
    UART_TXIE = UART_TXIF,
    UART_RCIP = UART_RCIF,
    UART_TXIP = UART_TXIF
#else
  #error "unsupported device " + _device
#endif  // _device
    
// alias public interrupt flags...
public dim
    RCIF    as UART_PIR.bits(UART_RCIF),    // receive buffer full
    TXIF    as UART_PIR.bits(UART_TXIF),    // transmit buffer empty
    RCIE    as UART_PIE.bits(UART_RCIE),    // receive interrupt enable
    TXIE    as UART_PIE.bits(UART_TXIE),    // transmit interrupt enable
    RCIP    as UART_IPR.bits(UART_RCIP),    // receive interrupt priority
    TXIP    as UART_IPR.bits(UART_TXIP)     // transmit interrupt priority

// public boolean flags...
public dim
    DataAvailable    as UART_PIR.booleans(UART_RCIF),   // RCIF
    ReadyToSend      as UART_PIR.booleans(UART_TXIF),   // TXIF
    RCIEnable        as UART_PIE.booleans(UART_RCIE),   // RCIE
    TXIEnable        as UART_PIE.booleans(UART_TXIE),   // TXIE
    RCIPHigh         as UART_IPR.booleans(UART_RCIP),   // RCIP
    TXIPHigh         as UART_IPR.booleans(UART_TXIP)    // TXIP

//--------------------------------------------------------------------------
// public baudrate constants
//--------------------------------------------------------------------------
// note: these are changed from the original SF libs
// there was an extra '+1' in the divisor which caused the rate to
// be way off for slow baudrates... only really an issue < 1200 baud
public const
    br300 as word    = FOSC / (FMULT * 300) - 1 + 0.5,
    br600 as word    = FOSC / (FMULT * 600) - 1 + 0.5,
    br1200 as word   = FOSC / (FMULT * 1200) - 1 + 0.5,
    br2400 as word   = FOSC / (FMULT * 2400) - 1 + 0.5,
    br4800 as word   = FOSC / (FMULT * 4800) - 1 + 0.5,
    br9600 as word   = FOSC / (FMULT * 9600) - 1 + 0.5,
    br19200 as word  = FOSC / (FMULT * 19200) - 1 + 0.5,
    br38400 as word  = FOSC / (FMULT * 38400) - 1 + 0.5,
    br57600 as word  = FOSC / (FMULT * 57600) - 1 + 0.5,
    br115200 as word = FOSC / (FMULT * 115200) - 1 + 0.5
#if (USART4_BRG16 and (_clock >= 16))
public const
    br230400 as word = FOSC / (FMULT * 230400) - 1 + 0.5,
    br250000 as word = FOSC / (FMULT * 250000) - 1 + 0.5,
    br460800 as word = FOSC / (FMULT * 460800) - 1 + 0.5
#endif

public dim
   ReadTerminator as char                    // read string terminator

{
****************************************************************************
* Name    : InitUSARTIO()                                                  *
* Purpose : Sets the hardware USART PIO pins                               *
* Note: this may not be required as enabling the usart typ sets the pin    *
* directions, but it is left here for backwards compatability and          *
* optional startup init via '#option USART4_STARTUP_INITIO'                *
* ** The user must setup any PPS functions **                              *
****************************************************************************
} 
public inline sub InitUSARTIO()
    input(USART_RX_PIO)         // receive pin is input
    high(USART_TX_PIO)          // transmit pin is output, idle high
end sub

{
****************************************************************************
* Name    : SetBaudrate                                                    *
* Purpose : Sets the hardware USART baudrate and USART mode registers      *
*         : Pass SPBRG constant, as defined above. For example, br115200   *
****************************************************************************
}
// baudrate parameter is passed as a word since SPBRG/SPBRGH may not be contiguous
#if (USART_XV18CORE)
public sub SetBaudrate(pSPBRG as word = br19200)
    InitUSARTIO()
    SPEN = 0                // disable usart while we setup regs
    
    UxCON0 = 0              // BRGS=0, TXEN=0, RXEN=0, MODE=0000 (async 8-bit UART mode)
    UxCON1 = 0              // port off
    UxCON2 = 0              // RUNOVF=0, RXPOL/TXPOL=0 (normal idle high), 1 stop, flow off
    UxERRIR = 0
    UxERRIR = 0
    UxUIR = 0
    UxFIFO = 0
    UxBRGL = pSPBRG.byte0
    UxBRGH = pSPBRG.byte1
  #if USART4_BRGH
    BRGS = 1
  #endif

    TXEN = 1                // enable transmitter
    RXEN = 1                // enable receiver
    SPEN = 1                // enable serial port
end sub
#else   // standard core device
public sub SetBaudrate(pSPBRG as word = br19200)
    InitUSARTIO()
    RCStatus = 0            // disable usart while we setup regs
    
    SPBRG_Reg = pSPBRG.byte0
  #if USART4_BRG16
    SPBRGH_Reg = pSPBRG.byte1
    BRG16 = 1
  #endif

  #if USART4_BRGH
    TXStatus = $24         // high speed BRG, enable TX
  #else
    TXStatus = $20         // low speed BRG, enable TX
  #endif
    CREN = 1                // enable continuous receive
    SPEN = 1                // enable serial port
end sub
#endif      // USART_XV18CORE

{
****************************************************************************
* Name    : Open                                                           *
* Purpose : Sets the hardware USART baudrate and USART mode registers      *
*         : Pass baudrate (NOT an SPBRG const, a real baudrate ie 115200)  *
*         : and current cpu clock freq, in MHz                             *
*         : This can take a LOT more code that the table method but lets   *
*         : you set non-standard rates and/or change clock freq at runtime *
****************************************************************************
}
public sub Open(baudrate as longword, clock_freq as byte=_clock)
    dim baud_div as word
    dim oscf as longword,
        lwdiv as longword

    // this function computes the BRG div using the equation
    //   BRG = FOSC / (FMULT * baudrate) + 0.5 - 1
    // numbers are scaled up by 2 so that the lsb represents 0.5
    clock_freq = clock_freq << 1    // scale up by 2 so we can round
    oscf = clock_freq * 1000000
    lwdiv =  baudrate * FMULT
    lwdiv = oscf/lwdiv
    lwdiv = lwdiv + 1               // plus 0.5 (to round)
    baud_div = word(lwdiv/2)        // scale back down (lwdiv >> 1, but less code)
    baud_div = baud_div - 1

    SetBaudrate(baud_div)
end sub

// disable uart
public sub Close()
    SPEN = 0
end sub

{
****************************************************************************
* Name    : ClearOverrun                                                   *
* Purpose : Clear USART overrun error by resetting receive flag            *
****************************************************************************
}
public macro mCLEAR_OVERRUN()
  #if (USART_XV18CORE)
    if (Overrun) then
        WREG = UxRXB        // read input buffer to clear a space in the fifo
        RXFOIF = 0          // clear overflow flag
    endif
  #else   // standard core device
    if (Overrun) then
        CREN = 0  // disable continuous receive
        CREN = 1  // enable continous receive
    endif
  #endif      // USART_XV18CORE
end macro

public sub ClearOverrun()
    mCLEAR_OVERRUN()
end sub

{
****************************************************************************
* Name    : macro REPEAT_UNTIL_BIT(reg, bitno)                             *
* Purpose : wait for a bit to be set                                       *
****************************************************************************
}
// get sfr access bank addr from device file (changed in v1.9)
const ACCESS_SFR_START = _access_sfr_start
const ACCESS_SFR_END = _access_sfr_end
const ACCESSBANK = _accessbank
// wait for a bit to be set (does not modify WREG)
macro REPEAT_UNTIL_BIT(reg, bitno)
  #if (WDT = true)
    asm-
    .if (reg >= ACCESS_SFR_START) && (reg <= ACCESS_SFR_END)
        clrwdt
        btfss reg, bitno     ; use access bank
        bra $-4
    .else
        banksel reg
        clrwdt
        btfss reg, bitno, 1  ; use BSR
        bra $-4
        movlb ACCESSBANK
    .endif
    end asm
  #else     // WDT = false
    asm-
    .if (reg >= ACCESS_SFR_START) && (reg <= ACCESS_SFR_END)
        btfss reg, bitno     ; use access bank
        bra $-2
    .else
        banksel reg
        btfss reg, bitno, 1  ; use BSR
        bra $-2
        movlb ACCESSBANK
    .endif
    end asm
  #endif
end macro

{
****************************************************************************
* Name    : TXEnable                                                       *
* Purpose : Enable/disable tx TXEN                                         *
****************************************************************************
}
public sub TXEnable(b as byte)
    while (TRMT = 0)        // wait for any current TX to finish
    end while
    if (b = 0) then         // disable TXEN (clears TXIF)
        TXEN = 0
    else                    // enable TXEN
        TXEN = 1
        while (TXIF = 0)    // wait for TXIF to re-assert
        end while
    endif        
end sub

{
****************************************************************************
* Name    : WriteByte                                                      *
* Purpose : Write a byte value to the hardware USART                       *
*         : Wait until ready to send is enabled, then send WREG byte       *
****************************************************************************
}
public sub WriteByte(pValue as WREG)
  #if (USART4_TXEN)
    PRODL = pValue          // save param
    if (TXEN = 0) then      // exit if TX is disabled
        exit
    endif        
    pValue = PRODL          // restore param
  #endif
   REPEAT_UNTIL_BIT(UART_PIR, UART_TXIF)
   TXRegister = WREG
end sub

{
****************************************************************************
* Name    : ReadByte                                                       *
* Purpose : Read a byte from the hardware USART                            *
*         : Waits for USART data, return byte in RCREGx                    *
****************************************************************************
}
public function ReadByte() as RCRegister
  #if (USART4_CLEAR_OERR)
    mCLEAR_OVERRUN()        // inline clear overrun
  #endif
    REPEAT_UNTIL_BIT(UART_PIR, UART_RCIF)
end function

{
****************************************************************************
* Name    : ReadBoolean                                                    *
* Purpose : Read a boolean from the hardware USART                         *
****************************************************************************
}
public function ReadBoolean() as boolean
   Result = boolean(ReadByte)
end function
{
****************************************************************************
* Name    : ReadWord                                                       *
* Purpose : Read a word from the hardware USART                            *
****************************************************************************
}
public function ReadWord() as word
   Result.Bytes(0) = ReadByte
   Result.Bytes(1) = ReadByte
end function
{
****************************************************************************
* Name    : ReadLongWord                                                   *
* Purpose : Read a long word from the hardware USART                       *
****************************************************************************
}
public function ReadLongWord() as longword
   Result.Bytes(0) = ReadByte
   Result.Bytes(1) = ReadByte
   Result.Bytes(2) = ReadByte
   Result.Bytes(3) = ReadByte
end function
{
****************************************************************************
* Name    : ReadFloat                                                      *
* Purpose : Read a float from the hardware USART                           *
****************************************************************************
}
public function ReadFloat() as float
   Result.Bytes(0) = ReadByte
   Result.Bytes(1) = ReadByte
   Result.Bytes(2) = ReadByte
   Result.Bytes(3) = ReadByte
end function

{
****************************************************************************
* Name    : WriteBoolean                                                   *
* Purpose : Write a boolean value to the hardware USART                    *
****************************************************************************
}
public sub WriteBoolean(pValue as boolean)
   WriteByte(byte(pValue))
end sub
{
****************************************************************************
* Name    : WriteWord                                                      *
* Purpose : Write a word value to the hardware USART                       *
****************************************************************************
}
public sub WriteWord(pValue as word)
   WriteByte(pValue.Bytes(0))
   WriteByte(pValue.Bytes(1))
end sub
{
****************************************************************************
* Name    : WriteLongWord                                                  *
* Purpose : Write a word value to the hardware USART                       *
****************************************************************************
}
public sub WriteLongWord(pValue as longword)
   WriteByte(pValue.Bytes(0))
   WriteByte(pValue.Bytes(1))
   WriteByte(pValue.Bytes(2))
   WriteByte(pValue.Bytes(3))
end sub
{
****************************************************************************
* Name    : WriteFloat                                                     *
* Purpose : Write a floating point number to the hardware USART            *
****************************************************************************
}
public sub WriteFloat(pValue as float)
   WriteByte(pValue.Bytes(0))
   WriteByte(pValue.Bytes(1))
   WriteByte(pValue.Bytes(2))
   WriteByte(pValue.Bytes(3))
end sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a boolean from the hardware USART                         *
****************************************************************************
}
sub ReadItem(byref pValue as boolean)
   pValue = boolean(ReadByte)
end sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a byte from the hardware USART                            *
****************************************************************************
}
sub ReadItem(byref pValue as byte)
   pValue = ReadByte
end sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a shortint from the hardware USART                        *
****************************************************************************
}
sub ReadItem(byref pValue as shortint)
   pValue = ReadByte
end sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a word from the hardware USART                            *
****************************************************************************
}
sub ReadItem(byref pValue as word)
   pValue = ReadWord
end sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read an integer from the hardware USART                        *
****************************************************************************
}
sub ReadItem(byref pValue as integer)
   pValue = ReadWord
end sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a long word from the hardware USART                       *
****************************************************************************
}
sub ReadItem(byref pValue as longword)
   pValue = ReadLongWord
end sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a long integer from the hardware USART                    *
****************************************************************************
}
sub ReadItem(byref pValue as longint)
   pValue = ReadLongWord
end sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a floating point number from the hardware USART           *
****************************************************************************
}
sub ReadItem(byref pValue as float)
   pValue = ReadFloat
end sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a string from the hardware USART. Use ReadTerminator      *
*         : to specify the input string terminator character.              *
*         : 1.5 Reworked for silicon error                                 *
****************************************************************************
}
sub ReadItem(byref pText as string)
   dim TextPtr as POSTINC0
   dim TextAddr as FSR0
   dim Value as byte

   TextAddr = addressof(pText)
   Value = ReadByte()
   while Value <> byte(ReadTerminator)
      TextPtr = Value
      Value = ReadByte()
   wend
   TextPtr = null
end sub
{
****************************************************************************
* Name    : Read (COMPOUND)                                                *
* Purpose : Read an item from the hardware USART                           *
****************************************************************************
}
public compound sub Read(ReadItem)
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a boolean value to the hardware USART                    *
****************************************************************************
}
sub WriteItem(pValue as boolean)
   WriteByte(byte(pValue))
end sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a byte value to the hardware USART                       *
****************************************************************************
}
sub WriteItem(pValue as WREG)
   WriteByte(pValue)
end sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a short int value to the hardware USART                  *
****************************************************************************
}
sub WriteItem(pValue as shortint)
   WriteByte(pValue)
end sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a word value to the hardware USART                       *
****************************************************************************
}
sub WriteItem(pValue as word)
   WriteWord(pValue)
end sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write an integer value to the hardware USART                   *
****************************************************************************
}
sub WriteItem(pValue as integer)
   WriteWord(pValue)
end sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a long word to the hardware USART                        *
****************************************************************************
}
sub WriteItem(pValue as longword)
   WriteLongWord(pValue)
end sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a long integer to the hardware USART                     *
****************************************************************************
}
sub WriteItem(pValue as longint)
   WriteLongWord(pValue)
end sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a floating point number to the hardware USART            *
****************************************************************************
}
sub WriteItem(pValue as float)
   WriteByte(pValue.Bytes(0))
   WriteByte(pValue.Bytes(1))
   WriteByte(pValue.Bytes(2))
   WriteByte(pValue.Bytes(3))
end sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a string value to the hardware USART                     *
****************************************************************************
}
// v1.6 - removed inline asm
sub WriteItem(pText as string)
    FSR0 = addressof(pText)
    while (INDF0 <> 0)
        WriteByte(POSTINC0)
    end while
end sub
{
****************************************************************************
* Name    : Write (COMPOUND)                                               *
* Purpose : Write an item to the hardware USART                            *
****************************************************************************
}
public compound sub Write(WriteItem)
{
****************************************************************************
* Name    : WaitFor                                                        *
* Purpose : Wait for a byte value to be received                           *
****************************************************************************
}
public function WaitFor(pValue as byte) as boolean
   result = ReadByte = pValue
end function
{
****************************************************************************
* Name    : WaitForTimeout                                                 *
* Purpose : Wait for a byte value to be received, with timeout value in    *
*         : milliseconds                                                   *
****************************************************************************
}
public function WaitForTimeout(pValue as byte, pTimeout as word) as boolean
   dim Counter as byte
   dim Timeout as word

   ClearOverrun()
   Result = false
   Counter = 10
   repeat
      Timeout = pTimeout
      while (Timeout > 0)
         if DataAvailable and (RCRegister = pValue) then
            Result = true
            exit
         endif
         delayus(100)
         dec(Timeout)
      wend
      dec(Counter)
   until Counter = 0
end function
{
****************************************************************************
* Name    : WaitForCount                                                   *
* Purpose : Wait for pAmount bytes to be received. The incoming data is    *
*         : stored in pArray                                               *
****************************************************************************
}
public sub WaitForCount(byref pArray() as byte, pCount as word)
   dim Index as word

   ClearOverrun()
   Index = 0
   while pCount > 0
      pArray(Index) = ReadByte()
      dec(pCount)
      inc(Index)
   wend
end sub
{
****************************************************************************
* Name    : WaitForStr                                                     *
* Purpose : Wait for a string to be received                               *
****************************************************************************
}
public sub WaitForStr(pStr as string)
   dim Index as byte

   ClearOverrun()
   Index = 0
   while pStr(Index) <> null
      if WaitFor(pStr(Index)) then
         inc(Index)
      else
         Index = 0
      endif
   wend
end sub
{
****************************************************************************
* Name    : WaitForStrCount                                                *
* Purpose : Wait for pAmount characters to be received. The incoming data  *
*         : is stored in pStr                                              *
****************************************************************************
}
public sub WaitForStrCount(byref pStr as string, pCount as word)
   dim Index as byte

   ClearOverrun()
   Index = 0
   while pCount > 0
      pStr(Index) = ReadByte()
      dec(pCount)
      inc(Index)
   wend
   pStr(Index) = null
end sub
{
****************************************************************************
* Name    : WaitForStrTimeout                                              *
* Purpose : Wait for a string to be received, with timeout value in        *
*         : milliseconds                                                   *
****************************************************************************
}
public function WaitForStrTimeout(pStr as string, pTimeout as word) as boolean
   dim Counter, StrIndex as byte
   dim Timeout as word
   dim Ch as byte

   ClearOverrun()
   Result = false
   StrIndex = 0
   Counter = 10
   repeat
      Timeout = pTimeout
      while Timeout > 0
         Ch = pStr(StrIndex)
         if Ch = 0 then
            result = true
            exit
         elseif DataAvailable then
            if Ch = RCRegister then
               inc(StrIndex)
            else
               StrIndex = 0
            endif
         endif
         delayus(100)
         dec(Timeout)
      wend
      dec(Counter)
   until Counter = 0
end function
{
****************************************************************************
* Name    : DataAvailableTimeout                                           *
* Purpose : Checks to see if a byte value has been received, with          *
*         : timeout in milliseconds                                        *
****************************************************************************
}
public function DataAvailableTimeout(pTimeout as word) as boolean
   dim Counter as byte
   dim Timeout as word

   ClearOverrun()
   Counter = 10
   repeat
      Timeout = pTimeout
      while (Timeout > 0) and not DataAvailable
         delayus(100)
         dec(Timeout)
      wend
      dec(Counter)
   until (Counter = 0) or DataAvailable
   result = DataAvailable
end function
{
****************************************************************************
* Name    : Rep                                                            *
* Purpose : Write a byte value, pAmount times                              *
****************************************************************************
}
public sub Rep(pValue, pAmount as byte)
   while pAmount > 0
      WriteByte(pValue)
      dec(pAmount)
   wend
end sub
{
****************************************************************************
* Name    : Skip                                                           *
* Purpose : Read pAmount bytes                                             *
****************************************************************************
}
public sub Skip(pAmount as byte)
   while pAmount > 0
      WREG = ReadByte()
      dec(pAmount)
   wend
end sub

//
// module initialisation...
//
ReadTerminator = null
#if (USART4_STARTUP_INITIO)
  InitUSARTIO()
#endif

end module
